#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <time.h>


#include "tcpserver.h"
#include "tcpendpoint.h"
#include "logger.h"


TcpServer::TcpServer( short _port, in_addr_t _addr )
{
    // keep parameters
    listenPort = _port;
    listenAddr = _addr;

    int ret, data;

    // create listen socket
    listenSock = socket(AF_INET, SOCK_STREAM, 0);
    if (listenSock < 0)
    {
        log( LOG_ERROR, "can't open socket" );
        return;
    }

    data = 1;
    ret = setsockopt( listenSock, SOL_SOCKET, SO_REUSEADDR, &data, sizeof(int));
    if ( ret < 0 )
    {
        log(LOG_ERROR, "can't set REUSADDR [%d:%s]", errno, (int)strerror(errno) );
        close( listenSock );
        listenSock = NULL;
        return;
    }

    struct sockaddr_in serv_addr;
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = listenAddr;
    serv_addr.sin_port = htons( listenPort );
    ret = bind( listenSock, (struct sockaddr *) &serv_addr, sizeof(serv_addr));
    if ( ret < 0 )
    {
        log(LOG_ERROR, "can't bind socket [%d:%s]", errno, (int)strerror(errno) );
        close( listenSock );
        listenSock = NULL;
        return;
    }
}

TcpServer::~TcpServer()
{
    if ( listenSock > 0 )
    {
        close( listenSock );
        listenSock = -1;
    }
}

int TcpServer::Run( int _backlog )
{
    int ret;
    timeval tv;

    fd_set masterfs, readfs;
    int fdmax;

    if ( listenSock < 0 )
        return -1;

    ret = listen( listenSock, _backlog );
    if ( ret < 0 )
    {
        log( LOG_ERROR, "error on listen [%d:%s]", errno, (int)strerror(errno) );
        return ret;
    }

    FD_ZERO( &masterfs );
    fdmax = listenSock;
    FD_SET( listenSock, &masterfs );

    exitRequest = false;

    while (!exitRequest)
    {
        // wait for event
        tv.tv_sec = 2;
        tv.tv_usec = 0;
        memcpy( &readfs, &masterfs, sizeof(masterfs) );
        ret = select( fdmax+1, &readfs, NULL, NULL, &tv );

        if ( ret < 0 )
        {
            log( LOG_ERROR, "select failed [%d:%s]", errno, (int)strerror(errno) );
            break;
        }
        else if ( ret == 0 )
        {
            // timeout, let's loop to check 'exitRequest'
        }
        else
        {
            if ( FD_ISSET( listenSock, &readfs ) )
            {
                // if event on listener, then a new connection occured
                TcpEndPoint *pEndPoint = Accept();
                if ( pEndPoint )
                {
                    FD_SET( pEndPoint->Sock(), &masterfs );
                    if ( pEndPoint->Sock() > fdmax )
                        fdmax = pEndPoint->Sock();
                    endpoints.push_back( pEndPoint );
                }
            }
            else
            {
                // event on endpoint socket, go through the list of endpoints
                std::list<TcpEndPoint*>::iterator itEP;
                for ( itEP=endpoints.begin() ; itEP != endpoints.end(); itEP++ )
                {
                    TcpEndPoint *pEndPoint = *itEP;
                    if ( (pEndPoint->Sock() > 0) && FD_ISSET( pEndPoint->Sock(), &readfs) )
                    {
                        ret = pEndPoint->Read();
                        // if error, close endpoint connection
                        if ( ret <= 0 )
                        {
                            FD_CLR( pEndPoint->Sock(), &masterfs );
                            delete pEndPoint;
                            endpoints.erase( itEP );
                        }
                        break;
                    }
                }
            }
        }
     } /* end of while */

    // stoppping all endpoints
    DBG1( "TcpServer::Run: deleting all remaining endpoints (n=%d)", endpoints.size() );
    std::list<TcpEndPoint*>::iterator itEP;
    for ( itEP=endpoints.begin() ; itEP != endpoints.end(); )
    {
        delete *itEP;
        itEP = endpoints.erase( itEP );
    }

    return ret;
}

TcpEndPoint *TcpServer::Accept( void )
{
    // new TcpEndPoint connection
    struct sockaddr_in endPointAddr;
    socklen_t addrlen = sizeof(endPointAddr);
    int endPointSock = accept( listenSock, (struct sockaddr *)&endPointAddr,&addrlen );

    if ( endPointSock < 0 )
    {
        log( LOG_ERROR, "accept TcpEndPoint failed [%d:%s]", errno, (int)strerror(errno) );
        return NULL;
    }
    // record new EndPoint
    TcpEndPoint *pEndPoint = new TcpEndPoint( endPointSock, endPointAddr );

    char strAddr[INET_ADDRSTRLEN+1];
    strcpy( strAddr, inet_ntoa( endPointAddr.sin_addr ) );
    log( LOG_INFO, "accepted TcpEndPoint from %s", (int)strAddr );

    return pEndPoint;
}
